(*--------------------------------------------------------------------------+
 |                                                                          |
 |                    Bote  outils pour les expressions                   |
 |                                                                          |
 |  M. Quercia, le 24/08/96                                                 |
 +--------------------------------------------------------------------------+*)


              (* +--------------------------------------------+
                 |  Mise en arbre d'une chane de caractres  |
                 +--------------------------------------------+ *)


#open "genlex";;
let lexer = make_lexer ["+"; "-"; "*"; "/"; "^"; "("; ")"];;


let rec expr p = function [< atome a; (suite_expr p a) b >] -> b

and atome = function
| [< 'Int n   >] -> Const(float_of_int n)
| [< 'Float n >] -> Const(n)
| [< 'Ident f; (var_ou_fct f) a >] -> a
| [< 'Kwd "-"; atome a >] -> CL [-1.0,a]
| [< 'Kwd "("; (expr 0) v; 'Kwd ")" >] -> v

and var_ou_fct f = function
| [< 'Kwd "("; (expr 0) v; 'Kwd ")" >] -> Fct(f,v)
| [< >] -> Var(f)

and suite_expr p a = function
| [< (infixe p) (q,f); (expr q) b; (suite_expr p (f a b)) c >] -> c
| [< >] -> a

and infixe p flux = match fst(stream_get flux) with
| Kwd "+" when 1 > p -> stream_next flux; (1,somme(1.0))
| Kwd "-" when 1 > p -> stream_next flux; (1,somme(-1.0))
| Kwd "*" when 2 > p -> stream_next flux; (2,produit(1.0))
| Kwd "/" when 2 > p -> stream_next flux; (2,produit(-1.0))
| Kwd "^" -> stream_next flux; (p+3,puiss)
| _ -> raise Parse_failure

and somme a x y = match x with
| CL(lx) -> CL(lx @ [a,y])
| _      -> CL[ 1.0,x; a,y ]

and produit a x y = match x,y with
| Const(cx),PG([b,z]) -> CL[cx,PG[a*.b,z]]
| Const(cx),_         -> CL[cx,if a = 1.0 then y else PG[a,y]]
| PG(lx),PG([b,z])    -> PG(lx @ [a*.b,z])
| PG(lx),_            -> PG(lx @ [a,y])
| _                   -> PG[ 1.0,x; a,y ]

and puiss x y = match y with
| Const(cy) -> PG[cy,x]
| _         -> Fct("exp",PG[1.0,y; 1.0,Fct("ln",x)])
;; 

let expression(s) =
  let i = ref (-1) in
  let ch() = incr i; if !i < string_length s then s.[!i] else raise Parse_failure in
  let st = lexer(stream_from ch) in
  try match st with [< (expr 0) v; end_of_stream _ >] -> v
  with exn -> prerr_endline("Erreur de syntaxe");
              prerr_endline(s);
              prerr_endline((make_string !i ` `) ^ "^");
              raise exn
;;

#close "genlex";;

                              (* +-------------+
                                 |  Affichage  |
                                 +-------------+ *)


#open "g_arbre";;

let rec garb e = match e with
| Const(c) -> G_noeud(string_of_float c,[])
| Var(v)   -> G_noeud(v,[])
| Fct(f,u) -> G_noeud(f,[garb u])
| CL(liste) -> G_noeud("CL",map garbCL liste)
| PG(liste) -> G_noeud("PG",map garbPG liste)

and garbCL(a,e) = G_noeud("*",[G_noeud(string_of_float a,[]); garb(e)])
and garbPG(a,e) = G_noeud("^",[garb(e); G_noeud(string_of_float a,[])])
;;

let print_exp e = print_g_arbre(garb e);;


                             (* +--------------+
                                |  Evaluation  |
                                +--------------+ *)

let feval f u =  match f with
| "exp"  -> Const(exp u)
| "ln"   -> Const(log u)
| "sqrt" -> Const(sqrt u)
| "sin"  -> Const(sin u)
| "cos"  -> Const(cos u)
| "tan"  -> Const(tan u)
| "asin" -> Const(asin u)
| "acos" -> Const(acos u)
| "atan" -> Const(atan u)
| _      -> Fct(f,Const(u))

and fder f u =  match f with
| "exp"  -> Fct("exp",u)
| "ln"   -> PG[-1.0,u]
| "sqrt" -> CL[0.5,PG[-0.5,u]]
| "sin"  -> Fct("cos",u)
| "cos"  -> CL[-1.0,Fct("sin",u)]
| "tan"  -> PG[-2.0,Fct("cos",u)]
| "asin" -> PG[-0.5,CL[1.0,Const 1.0; -1.0,PG[2.0,u]]]
| "acos" -> CL[-1.0,PG[-0.5,CL[1.0,Const 1.0; -1.0,PG[2.0,u]]]]
| "atan" -> PG[-1.0,CL[1.0,Const 1.0; 1.0,PG[2.0,u]]]
| _      -> Fct(f^"'",u)
;;


